-- (c) Copyright 2008 Rapid Deployment Software - See License.txt
--
--****
-- == Unit Testing Framework
--
-- <<LEVELTOC depth=2>>
--
-- === Background
-- Unit testing is the process of assuring that the smallest programming units
-- are actually delivering functionality that complies with their specification.
-- The units in question are usually individual routines rather than whole programs
-- or applications.
--
-- The theory is that if the components of a system are working correctly, then
-- there is a high probability that a system using those components can be made
-- to work correctly.
--
-- In Euphoria terms, this framework provides the tools to make testing and reporting on
-- functions and procedures easy and standardized. It gives us a simple way to
-- write a test case and to report on the findings.\\
-- Example~:
-- <eucode>
-- include std/unittest.e
--
-- test_equal( "Power function test #1", 4, power(2, 2))
-- test_equal( "Power function test #2", 4, power(16, 0.5))
--
-- test_report()
-- </eucode>
--
-- Name your test file in the special manner, ##t_NAME.e## and then simply run
-- ##eutest## in that directory.
--
-- {{{
-- C:\Euphoria> eutest
-- t_math.e:
--  failed: Bad math, expected: 100 but got: 8
--  2 tests run, 1 passed, 1 failed, 50.0% success
--
-- Test failure summary:
--    FAIL: t_math.e
--
-- 2 file(s) run 1 file(s) failed, 50.0% success--
-- }}}
--
-- In this example, we use the ##test_equal## function to record the result of
-- a test. The first parameter is the name of the test, which can be anything
-- and is displayed if the test fails. The second parameter is the expected
-- result - what we expect the function being tested to return. The third
-- parameter is the actual result returned by the function being tested. This
-- is usually written as a call to the function itself.
--
-- It is typical to provide as many test cases as would be required to give us
-- confidence that the function is being truly exercised. This includes calling
-- it with typical values and edge-case or exceptional values. It is also useful
-- to test the function's error handling by calling it with bad parameters.
--
-- When a test fails, the framework displays a message, showing the test's name,
-- the expected result and the actual result. You can configure the framework to
-- display each test run, regardless of whether it fails or not.
--
-- After running a series of tests, you can get a summary displayed by calling
-- the ##test_report()## procedure. To get a better feel for unit testing, have
-- a look at the provided test cases for the standard library in the //tests//
-- directory.

--****
-- === Constants
--

include std/io.e
include std/pretty.e
include std/search.e

--
-- Public Variables
--

-- Verbosity values
public enum
	TEST_QUIET = 0,
	TEST_SHOW_FAILED_ONLY,
	TEST_SHOW_ALL

--
-- Private variables
--

atom time_start = time(), time_test = time()

integer test_count = 0, tests_passed = 0, tests_failed = 0
sequence filename
integer verbose = TEST_SHOW_FAILED_ONLY

integer abort_on_fail = 0
integer wait_on_summary = 0
integer accumulate_on_summary = 0
integer logging = 0, log_fh = 0

--
-- Private utility functions
--

procedure add_log(object data)
	if log_fh = 0 then
		return
	end if

	puts(log_fh, "entry = ")
	pretty_print(log_fh, data, {2, 2, 1, 78, "%d", "%.15g"})
	puts(log_fh, "\n")
	flush(log_fh)
end procedure

procedure test_failed(sequence name, object a, object b)
	if verbose >= TEST_SHOW_FAILED_ONLY then
		printf(2, "  failed: %s, expected: ", {name})
		pretty_print(2, a, {2,2,1,78,"%d", "%.15g"})
		puts(2, " but got: ")
		pretty_print(2, b, {2,2,1,78,"%d", "%.15g"})
		puts(2, "\n")
	end if

	tests_failed += 1

	add_log({ "failed", name, a, b, time() - time_test })
	time_test = time()

	if abort_on_fail then
		puts(2, "Abort On Fail set.\n")
		abort(2)
	end if
end procedure

procedure test_passed(sequence name)
	if verbose >= TEST_SHOW_ALL then
		printf(2, "  passed: %s\n", {name})
	end if

	tests_passed += 1

	add_log({ "passed", name, time() - time_test })
	time_test = time()
end procedure

--
-- Global Testing Functions
--

--****
-- === Setup Routines

--**
-- Set the amount of information that is returned about passed and failed tests.
--
-- Parameters:
-- # ##verbosity##: an atom which takes predefined values for verbosity levels.
--
-- Comments:
-- The following values are allowable for ##verbosity##:
-- * TEST_QUIET = 0,
-- * TEST_SHOW_FAILED_ONLY = 1
-- * TEST_SHOW_ALL = 2
--
-- However, anything less than TEST_SHOW_FAILED_ONLY is treated as TEST_QUIET, and everything
-- above TEST_SHOW_ALL is treated as TEST_SHOW_ALL.
--
-- * At the lowest verbosity level, only the score is shown, ie the ratio passed tests/total tests.
-- * At the medium level, in addition, failed tests display their name, the expected outcome and
--   the outcome they got. This is the initial setting.
-- * At the highest level of verbosity, each test is reported as passed or failed.
--
-- If a file crashes when it should not, this event is reported no matter the verbosity level.
--
-- The command line switch ""-failed" causes verbosity to be set to medium at startup. The
-- command line switch ""-all" causes verbosity to be set to high at startup.
--
-- See Also:
-- [[:test_report]]

public procedure set_test_verbosity(atom verbosity)
	verbose = verbosity
end procedure

--**
-- Request the test report to pause before exiting.
--
-- Parameters:
--		# ##to_wait##: an integer, zero not to wait, nonzero to wait.
--
-- Comments:
-- Depending on the environment, the test results may be invisible if
-- ##set_wait_on_summary(1)## was not called prior, as this is not the default. The command
-- line switch "-wait" performs this call.
--
-- See Also:
-- [[:test_report]]

public procedure set_wait_on_summary(integer to_wait)
	wait_on_summary = to_wait
end procedure

--**
-- Request the test report to save run stats in "unittest.dat" before exiting.
--
-- Parameters:
--		# ##accumulate##: an integer, zero not to accumulate, nonzero to accumulate.
--
-- Comments:
-- The file "unittest.dat" is appened to with {t,f}\\
-- : where
-- :: //t// is total number of tests run
-- :: //f// is the total number of tests that failed
--

public procedure set_accumulate_summary(integer accumulate)
	accumulate_on_summary = accumulate
end procedure

--**
-- Set behavior on test failure, and return previous value.
--
-- Parameters:
--		# ##abort_test##: an integer, the new value for this setting.
--
-- Returns:
-- 		An **integer**, the previous value for the setting.
--
-- Comments:
-- By default, the tests go on even if a file crashed.

public function set_test_abort(integer abort_test)
	integer tmp = abort_on_fail
	abort_on_fail = abort_test

	return tmp
end function

--****
-- === Reporting

--**
-- Output test report
--
-- Comments:
--
-- The report components are described in the comments section for [[:set_test_verbosity]]. Everything
-- prints on the standard error device.
--
-- See Also:
--   [[:set_test_verbosity]]

public procedure test_report()
	atom score
	integer fh
	sequence fname

	if tests_failed > 0 or verbose >= TEST_SHOW_ALL then
		if test_count = 0 then
			score = 100
		else
			score = (tests_passed / test_count) * 100
		end if

		printf(2, "  %d tests run, %d passed, %d failed, %.1f%% success\n",
			{test_count, tests_passed, tests_failed, score})
	end if

	if accumulate_on_summary then
		fname = command_line()
		if equal(fname[1], fname[2]) then
			fname = fname[1]
		else
			fname = fname[2]
		end if
		fh = open("unittest.dat", "a")
		printf(fh, "{%d,%d,\"%s\"}\n", {test_count, tests_failed, fname})
		close(fh)
	end if

	add_log({ "summary", test_count, tests_failed, tests_passed, time() - time_start })

	if match("t_c_", filename) = 1 then
		puts(2, "  test should have failed but was a success\n")
		abort(0)
	else
		abort(tests_failed > 0)
	end if
end procedure

procedure record_result(integer success, sequence name, object a, object b)
	test_count += 1

	if success then
		test_passed(name)
	else
		test_failed(name, a, b)
	end if
end procedure

--****
-- === Tests
--

--**
-- Records whether a test passes by comparing two values.
--
-- Parameters:
--		# ##name##: a string, the name of the test
--		# ##expected##: an object, the expected outcome of some action
--		# ##outcome##: an object, some actual value that should equal the reference ##expected##.
--
-- Comments:
--
-- * For atoms, a fuzz of 1e-9 is used to assess equality.
-- * For sequences, no such fuzz is implemented.
--
-- A test is recorded as passed if equality holds between ##expected## and ##outcome##. The latter
-- is typically a function call, or a variable that was set by some prior action.
--
-- While ##expected## and ##outcome## are processed symmetrically, they are not recorded
-- symmetrically, so be careful to pass ##expected## before ##outcome## for better test failure
-- reports.
--
-- See Also:
-- [[:test_not_equal]], [[:test_true]], [[:test_false]], [[:test_pass]], [[test_fail]]

public procedure test_equal(sequence name, object expected, object outcome)
	integer success
	if sequence(expected) or sequence(outcome) then
		success = equal(expected,outcome)
	else
		if expected > outcome then
			success = ((expected-outcome) < 1e-9)
		else
			success = ((outcome-expected) < 1e-9)
		end if
	end if
	record_result(success, name, expected, outcome)
end procedure

--**
-- Records whether a test passes by comparing two values.
--
-- Parameters:
--		# ##name##: a string, the name of the test
--		# ##expected##: an object, the expected outcome of some action
--		# ##outcome##: an object, some actual value that should equal the reference ##expected##.
--
-- Comments:
-- * For atoms, a fuzz of 1e-9 is used to assess equality.
-- * For sequences, no such fuzz is implemented.
--
-- A test is recorded as passed if equality does not hold between ##expected## and ##outcome##. The
-- latter is typically a function call, or a variable that was set by some prior action.
--
-- See Also:
-- [[:test_equal]], [[:test_true]], [[:test_false]], [[:test_pass]], [[test_fail]]

public procedure test_not_equal(sequence name, object a, object b)
	integer success
	if sequence(a) or sequence(b) then
		success = not equal(a,b)
	else
		if a > b then
			success = ((a-b) >= 1e-9)
		else
			success = ((b-a) >= 1e-9)
		end if
	end if
	record_result(success, name, a, b)
end procedure

--**
-- Records whether a test passes by comparing two values.
--
-- Parameters:
--		# ##name##: a string, the name of the test
--		# ##outcome##: an object, some actual value that should not be zero.
--
-- Comments:
-- This assumes an expected value different from 0. No fuzz is applied when checking whether an
-- atom is zero or not. Use [[:test_equal]]() instead in this case.
--
-- See Also:
-- [[:test_equal]],  [[:test_not_equal]],[[:test_false]], [[:test_pass]], [[test_fail]]

public procedure test_true(sequence name, object outcome)
	integer success
	if sequence(outcome) then
		success = 0
	else
		success = (outcome != 0)
	end if
	record_result(success, name, 1, outcome )
end procedure

--**
-- Records whether a test passes by comparing two values.
--
-- Parameters:
--		# ##name##: a string, the name of the test
--		# ##outcome##: an object, some actual value that should be zero
--
-- Comments:
-- This assumes an expected value of 0. No fuzz is applied when checking whether an atom is zero
-- or not. Use [[:test_equal]]() instead in this case.
--
-- See Also:
-- [[:test_equal]],  [[:test_not_equal]],[[:test_true]], [[:test_pass]], [[test_fail]]

public procedure test_false(sequence name, object outcome)
	integer success
	if not integer(outcome) then
		success = 0
	else
		success = (outcome = 0)
	end if
	record_result(success, name, 0, outcome)
end procedure

--**
-- Records that a test failed.
--
-- Parameters:
--		# ##name##: a string, the name of the test
--
-- See Also:
-- [[:test_equal]],  [[:test_not_equal]],[[:test_true]], [[:test_false]], [[test_pass]]

public procedure test_fail(sequence name)
	record_result(0, name, 1, 0)
end procedure

--**
-- Records that a test passed.
--
-- Parameters:
--		# ##name##: a string, the name of the test
--
-- See Also:
-- [[:test_equal]],  [[:test_not_equal]],[[:test_true]], [[:test_false]], [[test_fail]]

public procedure test_pass(sequence name)
	record_result(1, name, 1, 1)
end procedure

sequence cmd = command_line()

filename = cmd[2]

for i = 3 to length(cmd) do
	if equal(cmd[i], "-all") then
		set_test_verbosity(TEST_SHOW_ALL)
	elsif equal(cmd[i], "-failed") then
		set_test_verbosity(TEST_SHOW_FAILED_ONLY)
	elsif equal(cmd[i], "-wait") then
		set_wait_on_summary(1)
	elsif begins(cmd[i], "-accumulate") then
		set_accumulate_summary(1)
	elsif equal(cmd[i], "-log") then
		log_fh = open("unittest.log", "a")
		add_log({"file", filename})
	end if
end for
